"""Unit tests for the argmerge._func::{parse_func, update_from_function} functions."""

import re

import pytest

from argmerge.env import ENV_PREFIX, parse_env
from tests.utils import no_error

# string pattern
# regex pattern
# neither string/regex pattern


@pytest.mark.parametrize(
    "env_mapping,threshold_kwargs,change_ledger,env_prefix,debug,expected_kwargs,"
    "expected_ledger,context,",
    [
        # No Environment Variables - no change
        (
            {},
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            "THRESH",
            False,
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            no_error,
        ),
        # No relevant Environment Variables
        (
            {"COLORTERM": "truecolor"},
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            "THRESH",
            False,
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            no_error,
        ),
        # Two Environment Variables - extractable values
        (
            {"THRESH_FIRST": "99", "THRESH_SECOND": "2"},
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            "THRESH",
            True,
            {"first": 99, "second": 2},
            {
                "first": {"label": "Environment Variable", "rank": 30},
                "second": {"label": "Environment Variable", "rank": 30},
            },
            no_error,
        ),
        # Three environment variables, one is a string
        (
            {
                "THRESH_FIRST": "99",
                "THRESH_SECOND": "2.0",
                "THRESH_THIRD": "laugh",
            },
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            "THRESH",
            True,
            {"first": 99, "second": 2, "third": "laugh"},
            {
                "first": {"label": "Environment Variable", "rank": 30},
                "second": {"label": "Environment Variable", "rank": 30},
                "third": {"label": "Environment Variable", "rank": 30},
            },
            no_error,
        ),
        # Use the environment variable to retrieve the prefix, PYTHRESH
        (
            {
                "NOT_RELEVANT_FIRST": "99",
                "THRESH_SECOND": "2.0",
                "THRESH_THIRD": "laugh",
            },
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            ENV_PREFIX,
            True,
            {"first": 1, "second": 2.0, "third": "laugh"},
            {
                "first": {"label": "Python Function default", "rank": 0},
                "second": {"label": "Environment Variable", "rank": 30},
                "third": {"label": "Environment Variable", "rank": 30},
            },
            no_error,
        ),
        # Do NOT Provide a capture group in the regex pattern
        # this will skip THRESH_B an THRESH_C
        # The proper Regex is r'^THRESH.([A-Z_-]+)$'
        (
            {
                "NOT_RELEVANT_FIRST": "99",
                "THRESH_SECOND": "2.0",
                "THRESH_THIRD": "laugh",
            },
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            re.compile("^THRESH.[A-Z_-]+"),
            True,
            {"first": 1},
            {
                "first": {"label": "Python Function default", "rank": 0},
            },
            no_error,
        ),
        (
            {
                "NOT_RELEVANT_FIRST": "99",
                "THRESH_SECOND": "2.0",
                "THRESH_THIRD": "laugh",
            },
            {"first": 1},
            {"first": {"label": "Python Function default", "rank": 0}},
            77,
            True,
            {"first": 1, "second": 2.0, "third": "laugh"},
            {
                "first": {"label": "Python Function default", "rank": 0},
                "second": {"label": "Environment Variable", "rank": 30},
                "third": {"label": "Environment Variable", "rank": 30},
            },
            pytest.raises(ValueError),
        ),
    ],
)
def test_parse_env(
    monkeypatch,
    env_mapping,
    threshold_kwargs,
    change_ledger,
    env_prefix,
    debug,
    expected_kwargs,
    expected_ledger,
    context,
):
    with monkeypatch.context() as m:
        for env_var_name, env_var_value in env_mapping.items():
            m.setenv(env_var_name, env_var_value)

        with context:
            threshold_kwargs_, change_ledger_ = parse_env(
                threshold_kwargs, change_ledger, env_prefix, debug
            )
            assert threshold_kwargs_ == expected_kwargs
            assert change_ledger_ == expected_ledger



from argmerge import threshold


@threshold(trace_level="DEBUG")
def main(first, second, third: float = 3.0, fourth: float = 4.0, fifth: int = 5):
    pass


if __name__ == "__main__":
    main(first=1, second="second")



# cython: linetrace=True
# distutils: define_macros=CYTHON_TRACE=1
import os
import re
import sys
from typing import Any

from loguru import logger as LOGGER

from argmerge.base import SourceParser
from argmerge.utils import extract_literals

__all__ = ["ENV_PREFIX", "parse_env"]

ENV_PREFIX = os.environ.get("PYTHRESH", "THRESH")


class EnvParser(SourceParser):
    """The parser the extracts relevant environment variables.

    Args:
        label (str): The debugging label to indicate an argument was set by environment
            variables.
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.
    """

    rank: int = 30
    label: str = "Environment Variable"

    def __call__(
        cls,
        threshold_kwargs: dict[str, Any],
        change_ledger: dict[str, dict[str, str | int]],
        env_prefix: str | re.Pattern[str] = ENV_PREFIX,
        debug=False,
        **kwargs,
    ):
        """Parse the environment variables using the `env_prefix` and update inputs.

        Args:
            threshold_kwargs (dict[str, Any]): kwargs passed around the
                @threshold decorator.
            change_ledger (dict[str, dict[str, str  |  int]]): Tracks when kwargs are
                updated inside the @threshold decorator.
            env_prefix (str | re.Pattern[str], optional): The prefix used to search for
                set environment variables. Defaults to ENV_PREFIX, which is 'THRESH_'.
            debug (bool, optional): Flag to turn on more logging. Defaults to False.

        Raises:
            ValueError: `env_prefix` must either be a string or Regex string pattern.

        Returns:
            tuple[dict, dict]: an updated `threshold_kwargs` and `change_ledger`.
        """
        if debug:
            LOGGER.remove()
            LOGGER.add(sys.stderr, level="DEBUG")

        if isinstance(env_prefix, re.Pattern):
            pattern = env_prefix

        elif isinstance(env_prefix, str):
            pattern = re.compile(rf"(?:{env_prefix.upper()}.)([A-Za-z0\-\_]+)")

        else:
            raise ValueError(
                f"'env_prefix' must be either a string or Regex string pattern. Received: {env_prefix} ({type(env_prefix)})."
            )

        LOGGER.debug(f"{env_prefix=}")
        LOGGER.debug(f"{pattern=}")

        _env_kwargs = {}

        for k, v in os.environ.items():
            _search = pattern.search(k)

            if _search is not None:
                try:
                    key = _search.group(1).lower()
                    LOGGER.debug(f"{key=} {v=}")
                    _env_kwargs[key] = extract_literals(v)

                except IndexError:
                    LOGGER.debug(f"Regex search failed on environment variable {k}.")

            else:
                LOGGER.debug(f"Miss: {k=}")

        LOGGER.debug(f"{_env_kwargs=}")
        threshold_kwargs.update(_env_kwargs)

        LOGGER.debug(f"Updated {threshold_kwargs=}")

        for key in _env_kwargs:
            change_ledger[key] = {"label": cls.label, "rank": cls.rank}

        return threshold_kwargs, change_ledger


# Make EnvParser appear as a function when it uses __call__.
parse_env = EnvParser()



import pytest

from argmerge.base import SourceParser


def test_empty(monkeypatch):
    monkeypatch.setattr(SourceParser, "__abstractmethods__", set())
    SourceParser()
    # did this for coverage


def test_bad_class_declaration():
    with pytest.raises(TypeError):

        class Bad(SourceParser):
            label: str = "Bad"
            rank = int = -100

        Bad()



"""Module for utility functions."""

import ast

from loguru import logger as LOGGER


def extract_literals(s: str):
    try:
        return ast.literal_eval(s)
    except Exception as e:
        LOGGER.debug(e)
        return s



import pytest

from argmerge.cli import CLI_PATTERN, parse_cli
from tests.utils import no_error

test_cases = [
    (
        ["", "--config=config.txt", "--debug=True"],
        {},
        {},
        CLI_PATTERN,
        True,
        {"config": "config.txt", "debug": "True"},
        {"config": {"label": "CLI", "rank": 40}, "debug": {"label": "CLI", "rank": 40}},
        no_error,
    ),
    (
        ["", "--config", "config.txt", "--debug=True"],
        {},
        {},
        r"--([A-Za-z\_\-]+)\=([0-9A-Za-z\_\-\.]+)",
        True,
        {"debug": "True"},
        {"debug": {"label": "CLI", "rank": 40}},
        no_error,
    ),
]


@pytest.mark.parametrize(
    "cli_args,threshold_kwargs,change_ledger,cli_pattern,"
    "debug,expected_kwargs,expected_ledger,context",
    test_cases,
)
def test_cli_parsing(
    monkeypatch,
    cli_args,
    threshold_kwargs,
    change_ledger,
    cli_pattern,
    debug,
    expected_kwargs,
    expected_ledger,
    context,
):
    monkeypatch.setattr("sys.argv", cli_args)

    with context:
        threshold_kwargs_, change_ledger_ = parse_cli(
            threshold_kwargs, change_ledger, cli_pattern, debug
        )

        assert threshold_kwargs_ == expected_kwargs
        assert change_ledger_ == expected_ledger



"""Module that contains the main decorator, `@threshold`.

Usage:
    ```py
    # main.py
    from argmerge import threshold

    @threshold
    def main(first: int, second: str, third: float = 3.0):
        ...

    if __name__ == '__main__':
        main()
    ```
    Many more examples of how to use this can be found in [the Examples section](/examples/)
"""

import functools
import re
from pathlib import Path

from argmerge.cli import CLI_PATTERN, parse_cli
from argmerge.env import ENV_PREFIX, parse_env
from argmerge.func import parse_func, parse_func_runtime
from argmerge.json import parse_json
from argmerge.trace import LOG_LEVELS, trace_arg_lineage
from argmerge.yaml import parse_yaml


def threshold(
    *args,
    fpath_json: str | Path = "",
    fpath_yaml: str | Path = "",
    env_prefix: str | re.Pattern[str] = ENV_PREFIX,  # 'THRESH', also set at PYTHRESH
    cli_pattern: str | re.Pattern[str] = CLI_PATTERN,
    trace_level: str = "",
    debug: bool = False,
    **kwargs,
):
    """Merge arguments from external environment sources into the program.

    We allow syntax of both @threshold and @threshold(), depending whether you want to
    allow for defaults or override them.

    Args:
        fpath_json (str | Path, optional): The path to find a JSON configuration file.
            Defaults to "".
        fpath_yaml (str | Path, optional): The path to find a YAML configuration file.
            Defaults to "".
        env_prefix (str | re.Pattern[str], optional): The string or Regex to match
            environment variables against. Defaults to ENV_PREFIX, which is 'THRESH'.
        cli_pattern (str | re.Pattern[str], optional): The string or Regex to match
            CLI arguments against. Defaults to CLI_PATTERN.
        trace_level (str, optional): Trace the source of each kwarg and display at the
            specified trace log level. Defaults to "", which skips the trace entirely.
        debug (bool, optional): Turns on debugging for all the parsers.
            Defaults to False.

    Raises:
        ValueError: `level` must be in one of the default loguru logger levels:
            `("critical", "warning", "success", "info", "debug")`

    Returns:
        callable: A wrapped, yet-to-be-called function with resolved arguments.
    """
    if len(args) == 1:
        # allow syntax of @threshold and @threshold()
        return threshold()(args[0])

    else:

        def wrapped(f):
            @functools.wraps(f)
            def wrapped_f(*_args, **_kwargs):
                _threshold_kwargs, _change_ledger = dict(), dict()
                _threshold_kwargs, _change_ledger = parse_func(
                    _threshold_kwargs, _change_ledger, f, debug=debug
                )

                _threshold_kwargs, _change_ledger = parse_json(
                    _threshold_kwargs,
                    _change_ledger,
                    fpath_json=fpath_json,
                    debug=debug,
                )

                _threshold_kwargs, _change_ledger = parse_yaml(
                    _threshold_kwargs,
                    _change_ledger,
                    fpath_yaml=fpath_yaml,
                    debug=debug,
                )

                _threshold_kwargs, _change_ledger = parse_env(
                    _threshold_kwargs,
                    _change_ledger,
                    env_prefix=env_prefix,
                    debug=debug,
                )

                _threshold_kwargs, _change_ledger = parse_cli(
                    _threshold_kwargs,
                    _change_ledger,
                    cli_pattern=cli_pattern,
                    debug=debug,
                )

                _threshold_kwargs, _change_ledger = parse_func_runtime(
                    _threshold_kwargs, _change_ledger, func_kwargs=_kwargs, debug=debug
                )

                if trace_level.lower() in LOG_LEVELS:
                    trace_arg_lineage(
                        f,
                        _change_ledger,
                        level=trace_level,
                    )

                elif trace_level == "":
                    # default behavior
                    pass

                else:
                    raise ValueError(
                        f"'trace_level' has been set to '{trace_level}', which is not "
                        "supported. Please set 'trace_level' to an empty string or one"
                        f" of: {LOG_LEVELS}."
                    )

                return f(*_args, **_threshold_kwargs)

            return wrapped_f

        return wrapped



"""Module to write the source of each function keyword argument for the developer

Example:

    $ uv run main.py  # trace_level=DEBUG set in the program
    2025-10-14 16:15:33.466 | DEBUG    | argmerge.trace:_write_trace:50 -
    Parameter Name  | Source
    =======================================
    third   | Python Function default
    fourth  | Python Function default
    fifth   | Python Function default
    first   | developer-provided
    second  | developer-provided
    =======================================
"""

import sys
from inspect import signature
from typing import Callable

from loguru import logger as LOGGER

LOG_LEVELS = ("critical", "warning", "success", "info", "debug")


def _write_trace(message: str, level: str):
    """Write the trace message out to the specified log level.

    args:
        message (str): The debugging message for the developer.
        level (str): The `loguru` log level to set.

    raises:
        TypeError: `level` must be a string.
        ValueError: `level` must be in one of the default loguru logger levels:
            `("critical", "warning", "success", "info", "debug")`
    """
    if isinstance(level, str):
        _level = level.lower()

    else:
        raise TypeError(f"Log level '{level}' ({type(level)}) is not a string.")

    if _level not in LOG_LEVELS:
        raise ValueError(
            f"Log level '{_level}' not in basic loguru log levels: '{LOG_LEVELS}''."
        )

    _trace_logger: int = LOGGER.add(sys.stderr, level=level.upper())

    getattr(LOGGER, _level.lower())(message)

    LOGGER.remove(_trace_logger)


def _log_trace(ledger: dict[str, dict[str, str | int]], level: str = ""):
    """Compose a developer-friendly report of each kwarg's source form the ledger.

    First, sort the keys by their ranks. Ranks are sorted in ascending order such that
    the highest ranked sources (Developer-Provided, etc) will be seen first. Then sort
    the labels by that same order. Next, we will dynamically produce the report. We
    will find the kwarg and source each with the longest length. Then, left-justify
    (ljust) each kwarg and source to fill the space. Repeat for each kwarg-source pair.
    We then set the heading such that the '=' characters "hang" over the report,
    visually grouping them.

    Args:
        ledger (dict[str, dict[str, str  |  int]]): The change ledger that describes
            each kwarg's highest-ranked source.
        level (str, optional): A `loguru` logging level. Defaults to "", which means
            'DEBUG' will be set.
    """

    if len(ledger) == 0:
        LOGGER.warning("Change ledger is empty, will not write out!")

    else:
        # split the ranks from the labels
        ranks: dict[str, int] = {k: int(v["rank"]) for k, v in ledger.items()}
        labels: dict[str, str] = {k: str(v["label"]) for k, v in ledger.items()}

        # First, sort the keys by their ranks.
        sorted_keys = [x for x, _ in sorted(ranks.items(), key=lambda x: x[1])]
        sorted_labels = {k: labels[k] for k in sorted_keys}

        # calculate the longest kwarg name and source name
        _key_spacing: int = max(list(map(len, sorted_labels.keys())))
        _value_spacing: int = max(list(map(len, sorted_labels.values())))

        # left-justify
        _pre_join_spacing = {
            k.ljust(_key_spacing, " "): v.ljust(_value_spacing, " ")
            for k, v in sorted_labels.items()
        }

        # stringified, left-justified kwargs and sources
        _params = [f"{k}\t| {v}" for k, v in _pre_join_spacing.items()]

        # set up the heading
        # heading will extend over the params
        _heading_spacing: int = max(map(len, _params)) + 7

        _heading_param = "Parameter Name".ljust(_key_spacing)
        _heading_loc = "Source".ljust(_value_spacing)

        # construct the full heading
        _heading = f"{_heading_param}\t| {_heading_loc}"

        _body = "\n".join(_params)

        # combine the heading with the body
        msg = (
            f"\n{_heading}\n{'=' * _heading_spacing}\n{_body}\n{'=' * _heading_spacing}"
        )

        # log it to the appropriate level
        _write_trace(message=msg, level=level)


def trace_arg_lineage(
    f: Callable,
    change_ledger: dict[str, dict[str, str | int]],
    level: str = "",
):
    """Determine where each argument in the function came from.

    Only include arguments that exist in the function header. If a function accepts
    **kwargs and an irrelevant keyword is provided - discard it.

    args:
        f (callable):
        change_ledger (dict): The final dictionary detailing where every argument
            is set - defaults, files, environment variables, CLI arguments, etc.
    """
    sig = signature(f)
    _changed = {k: v for k, v in change_ledger.items() if k in sig.parameters}

    _log_trace(ledger=_changed, level=level)



"""Module that holds the Abstract Base Class for all external parsers."""

from abc import ABC, abstractmethod

__all__ = ["SourceParser"]


class SourceParser(ABC):
    """The base parser from which all other parsers are built.

    To build a new parser, you will need to subclass this one. Set 'rank' and 'label'
    as class variables and overwrite `__call__` with your parser. You MUST include
    three parameters: `threshold_kwargs`, `ledger`, and `debug`. This will allow you
    to add your own with other keyword arguments. You can look at any of the existing
    subclasses for how this works. After you build your subclass, you will instantiate
    it, setting it equal to a variable, ideally beginning with `'parse_'`. This will
    allow you to treat it like a function by invoking the `__call__` method, which calls
    your parser.

    Args:
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.
        label (str): The debugging label to indicate an argument was set at the
            <source level>.
    """

    rank: int = -100
    label: str = ""

    def __init__(cls):
        cls.label
        cls.rank

    @abstractmethod
    def __call__(
        cls, threshold_kwargs: dict, ledger: dict, debug: bool = False, **kwargs
    ) -> tuple[dict, dict]:
        """This is too abstract to be covered"""



import json
import sys
from pathlib import Path
from typing import Any

from loguru import logger as LOGGER

from argmerge.base import SourceParser

__all__ = ["parse_json"]


class JSONParser(SourceParser):
    """The parser the extracts relevant arguments from a JSON file.

    params:
        label (str): The debugging label to indicate an argument was set in a JSON
            config file.
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.

    """

    label: str = "JSON"
    rank: int = 10

    def __call__(
        cls,
        threshold_kwargs: dict[str, Any],
        change_ledger: dict[str, dict[str, str | int]],
        fpath_json: str | Path,
        debug: bool = False,
    ) -> tuple[dict, dict]:
        """Parse a JSON configuration file for arguments

         Args:
            threshold_kwargs (dict[str, Any]): kwargs passed around the
                @threshold decorator.
            change_ledger (dict[str, dict[str, str  |  int]]): Tracks when kwargs are
                updated inside the @threshold decorator.
            fpath_json (str | Path): The filepath to the JSON configuration file.
            debug (bool, optional): Flag to turn on more logging. Defaults to False.

        Raises:
            ValueError: If filepath extension is not `json`.

        Returns:
            tuple[dict, dict]: an updated `threshold_kwargs` and `change_ledger`.
        """
        _json_kwargs: dict

        if debug:
            LOGGER.remove()
            LOGGER.add(sys.stderr, level="DEBUG")

        LOGGER.debug(f"{threshold_kwargs=}")
        LOGGER.debug(f"{fpath_json=}")

        _fpath_json = Path(fpath_json)
        if _fpath_json == Path(""):
            LOGGER.debug("fpath_json not provided, skipping.")

        else:
            if _fpath_json.suffix != ".json":
                raise ValueError(
                    f"The JSON suffix of '{_fpath_json.as_posix()}' is not correct."
                    " Please use '.json'."
                )

            cls.label = f"JSON ({_fpath_json})"

            with open(fpath_json, "rb") as fy:
                _json_kwargs = json.load(fy)

            LOGGER.debug(f"{_json_kwargs=}")
            threshold_kwargs.update(_json_kwargs)
            LOGGER.debug(f"Updated {threshold_kwargs=}")

            for key in _json_kwargs:
                change_ledger[key] = {"label": cls.label, "rank": cls.rank}

        return threshold_kwargs, change_ledger


parse_json = JSONParser()



"""Unit tests for the argmerge._json::parse_json() function."""

import json

import pytest

from argmerge.json import parse_json
from tests.utils import no_error


@pytest.mark.parametrize(
    "threshold_kwargs,change_ledger,fpath_json,debug,expected_kwargs,expected_ledger,context",
    [
        # file does not exist
        pytest.param(
            {},
            {},
            "missing_config.json",
            False,
            None,
            None,
            pytest.raises(FileNotFoundError),
        ),
        # wrong filepath extension -> ValueError
        pytest.param(
            {},
            {},
            "fpath_config.cfg",
            False,
            None,
            None,
            pytest.raises(ValueError),
        ),
        # broken JSON -> error
        pytest.param(
            {},
            {},
            "tests/json/bad_config.json",
            False,
            None,
            None,
            pytest.raises(json.JSONDecodeError),
        ),
        # working example
        # debug true
        pytest.param(
            {},
            {},
            "tests/json/good_config.json",
            True,
            {"first": 1, "second": {"third": 3, "d": 4}},
            {
                "first": {"label": "JSON (tests/json/good_config.json)", "rank": 10},
                "second": {"label": "JSON (tests/json/good_config.json)", "rank": 10},
            },
            no_error,
        ),
        # debug false
        pytest.param(
            {"first": 7, "third": 9},
            {
                "first": {"label": "JSON (other.json)", "rank": 10},
                "third": {"label": "JSON (other.json)", "rank": 10},
            },
            "tests/json/good_config.json",
            True,
            {"first": 1, "second": {"third": 3, "d": 4}, "third": 9},
            {
                "first": {"label": "JSON (tests/json/good_config.json)", "rank": 10},
                "second": {"label": "JSON (tests/json/good_config.json)", "rank": 10},
                "third": {"label": "JSON (other.json)", "rank": 10},
            },
            no_error,
        ),
    ],
)
def test_parse_json(
    threshold_kwargs,
    change_ledger,
    fpath_json,
    debug,
    expected_kwargs,  # output[0]
    expected_ledger,  # output[1]
    context,  # whether or not is an error
):
    with context:
        threshold_kwargs_, change_ledger_ = parse_json(
            threshold_kwargs, change_ledger, fpath_json, debug=debug
        )
        assert threshold_kwargs_ == expected_kwargs
        assert change_ledger_ == expected_ledger



"""Module to work with retrieving arguments functions"""

import sys
from inspect import Parameter, signature
from typing import Any, Callable

from loguru import logger as LOGGER

from argmerge.base import SourceParser

__all__ = ["parse_func", "parse_func_runtime"]


class FuncDefaultParser(SourceParser):
    """Builds the parser to extract default arguments from a function signature

    This is the lowest-ranked parser.

    params:
        label (str): The debugging label to indicate an argument was set at the CLI.
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.

    """

    label: str = "Python Function Default"
    rank: int = 0

    def __call__(
        cls,
        threshold_kwargs: dict,
        change_ledger: dict,
        f: Callable,
        debug: bool = False,
    ) -> tuple[dict, dict]:
        """Lowest level parser - retrieve function defaults as fallback arguments.

        Args:
            threshold_kwargs (dict[str, Any]): kwargs passed around the
                @threshold decorator.
            change_ledger (dict[str, dict[str, str  |  int]]): Tracks when kwargs are
                updated inside the @threshold decorator.
            f (Callable): The function we wrap.
            debug (bool, optional): Flag to turn on more logging. Defaults to False.


        Returns:
            tuple[dict, dict]: an updated `threshold_kwargs` and `change_ledger`.
        """

        if debug:
            LOGGER.remove()
            LOGGER.add(sys.stderr, level="DEBUG")

        _sig = signature(f)
        LOGGER.debug(f"Function {signature=}")
        _default: dict = {}

        for k, v in _sig.parameters.items():
            if v.default is not Parameter.empty:
                _default[k] = v.default

        LOGGER.debug(f"{_default=}")
        for k in _default:
            change_ledger[k] = {"label": cls.label, "rank": cls.rank}

        threshold_kwargs.update(**_default)

        return threshold_kwargs, change_ledger


parse_func = FuncDefaultParser()


class FuncUpdater(SourceParser):
    """Builds the parser to extract default arguments from a function signature

    This is the highest-ranked parser.

    params:
        label (str): The debugging label to indicate an argument was set at Runtime by
            the developer.
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.
    """

    label: str = "Developer-provided"
    rank: int = 100

    def __call__(
        cls,
        threshold_kwargs: dict[str, Any],
        change_ledger: dict[str, dict[str, str | int]],
        func_kwargs: dict[str, Any],
        debug: bool = False,
    ) -> tuple[dict, dict]:
        """Update the external values with the function's runtime arguments.

        Args:
            threshold_kwargs (dict[str, Any]): kwargs passed around the
                @threshold decorator.
            change_ledger (dict[str, dict[str, str  |  int]]): Tracks when kwargs are
                updated inside the @threshold decorator.
            func_kwargs (dict[str, Any]): The Runtime kwargs of the function.
            debug (bool, optional): Flag to turn on more logging. Defaults to False.

        Returns:
            Returns:
            tuple[dict, dict]: an updated `threshold_kwargs` and `change_ledger`.
        """
        if debug:
            LOGGER.remove()
            LOGGER.add(sys.stderr, level="DEBUG")

        LOGGER.debug(f"{threshold_kwargs=}")
        LOGGER.debug(f"{func_kwargs=}")

        threshold_kwargs.update(**func_kwargs)

        for key in func_kwargs:
            change_ledger[key] = {"label": cls.label, "rank": cls.rank}

        return threshold_kwargs, change_ledger


parse_func_runtime = FuncUpdater()



"""Module that provides a flexible CLI parser component in the decorator.


```py
CLI_PATTERN: re.Pattern = re.compile(r"--([A-Za-z_-]+)=([0-9A-Za-z._-]+)")
```

- matches `'--arg=value'`
- does not match `'--arg value'`
"""

import re
import sys
from typing import Any

from loguru import logger as LOGGER

from argmerge.base import SourceParser

__all__ = ["CLI_PATTERN", "parse_cli"]

# matches '--arg=value'
# does not match '--arg value'
CLI_PATTERN: re.Pattern = re.compile(r"--([A-Za-z_-]+)=([0-9A-Za-z._-]+)")


class CLIParser(SourceParser):
    """The parser the extracts relevant CLI arguments.

    params:
        label (str): The debugging label to indicate an argument was set at the CLI.
        rank (int): The priority of the parser. Generally, we aim between [0,100] for
            human-readabilty.

    """

    label: str = "CLI"
    rank: int = 40

    def __call__(
        cls,
        threshold_kwargs: dict[str, Any],
        change_ledger: dict[str, dict[str, str | int]],
        cli_pattern: re.Pattern[str] = CLI_PATTERN,
        debug: bool = False,
        **kwargs,
    ) -> tuple[dict, dict]:
        """Parse the CLI commands using the cli_pattern regular expression.

        Args:
            threshold_kwargs (dict[str, Any]): kwargs passed around the
                @threshold decorator.
            change_ledger (dict[str, dict[str, str  |  int]]): Tracks when kwargs are
                updated inside the @threshold decorator.
            cli_pattern (re.Pattern[str], optional): The regular expression pattern
                used to extract arguments from the CLI. Defaults to CLI_PATTERN.
            debug (bool, optional): Flag to turn on more logging. Defaults to False.

        Returns:
            tuple[dict, dict]: an updated `threshold_kwargs` and `change_ledger`.
        """
        _cli_kwargs: dict
        _cli_input: str

        if debug:
            LOGGER.remove()
            LOGGER.add(sys.stderr, level="DEBUG")

        if isinstance(cli_pattern, re.Pattern):
            _cli_pattern = cli_pattern

        else:
            _cli_pattern = re.compile(rf"{cli_pattern}")

        LOGGER.debug(f"{cli_pattern=}")
        LOGGER.debug(f"{_cli_pattern=}")
        LOGGER.debug(f"{sys.argv=}")
        _cli_input = " ".join(sys.argv[1:])
        LOGGER.debug(f"{_cli_input}")

        _cli_kwargs = dict(_cli_pattern.findall(_cli_input))
        LOGGER.debug(f"{_cli_kwargs=}")

        threshold_kwargs.update(_cli_kwargs)
        LOGGER.debug(f"Updated {threshold_kwargs=}")

        for key in _cli_kwargs:
            change_ledger[key] = {"label": cls.label, "rank": cls.rank}

        return threshold_kwargs, change_ledger


parse_cli = CLIParser()



"""Unit tests for the argmerge._func::{parse_func, parse_func_runtime} functions."""

import pytest

from argmerge.func import parse_func, parse_func_runtime
from tests.utils import no_error


# [parse_func]
def no_default_params(first: int, second: str, third: float):
    pass


def mixed_params(first: int, second: str, third: float = 0.0):
    pass


def all_default_params(first: int = 7, second: str = "", third: float = 0.0):
    pass


@pytest.mark.parametrize(
    "threshold_kwargs, change_ledger,function, debug,expected_kwargs, expected_ledger, context",
    [
        # if not a callable is passed
        (None, None, None, False, None, None, pytest.raises(TypeError)),
        # if no parameters with defaults are passed
        ({}, {}, no_default_params, False, {}, {}, no_error),
        # if mix of default/not default params are passed
        (
            {},
            {},
            mixed_params,
            True,
            {"third": 0.0},
            {"third": {"label": "Python Function Default", "rank": 0}},
            no_error,
        ),
        # if all functions have defaults
        (
            {},
            {},
            all_default_params,
            True,
            {"first": 7, "second": "", "third": 0.0},
            {
                "third": {"label": "Python Function Default", "rank": 0},
                "second": {"label": "Python Function Default", "rank": 0},
                "first": {"label": "Python Function Default", "rank": 0},
            },
            no_error,
        ),
    ],
)
def test_parse_func(
    function,
    debug,
    threshold_kwargs,
    change_ledger,
    expected_kwargs,
    expected_ledger,
    context,
):
    with context:
        _expected_kwargs, _expected_ledger = parse_func(
            threshold_kwargs, change_ledger, function, debug
        )
        assert expected_kwargs == _expected_kwargs
        assert expected_ledger == _expected_ledger


# [parse_func_runtime]
@pytest.mark.parametrize(
    "threshold_kwargs,change_ledger,func_kwargs,debug,expected_kwargs,expected_ledger,context,",
    [
        # no previous changes
        (
            {},
            {},
            {"first": 1},
            False,
            {"first": 1},
            {"first": {"label": "Developer-provided", "rank": 100}},
            no_error,
        ),
        # one previous change
        (
            {"first": 3},
            {"first": {"label": "Python Function Default", "rank": 0}},
            {"first": 1},
            False,
            {"first": 1},
            {"first": {"label": "Developer-provided", "rank": 100}},
            no_error,
        ),
        # many previous changes from many levels
        (
            {"first": 1, "second": "", "third": 0.0},
            {
                "first": {"label": "Python Function Default", "rank": 0},
                "second": {"label": "Environment Variable", "rank": 30},
                "third": {"label": "CLI", "rank": 40},
            },
            {"first": 99, "second": "Ken Thompson", "third": -0.8},
            True,
            {"first": 99, "second": "Ken Thompson", "third": -0.8},
            {
                "first": {"label": "Developer-provided", "rank": 100},
                "second": {"label": "Developer-provided", "rank": 100},
                "third": {"label": "Developer-provided", "rank": 100},
            },
            no_error,
        ),
    ],
)
def test_parse_func_runtime(
    threshold_kwargs,
    change_ledger,
    func_kwargs,
    debug,
    expected_kwargs,
    expected_ledger,
    context,
):
    with context:
        _expected_kwargs, _expected_ledger = parse_func_runtime(
            threshold_kwargs, change_ledger, func_kwargs, debug
        )
        assert expected_kwargs == _expected_kwargs
        assert expected_ledger == _expected_ledger



import pytest

from argmerge import threshold
from argmerge.cli import CLI_PATTERN
from argmerge.env import ENV_PREFIX
from tests.utils import no_error

EMPTY_TEST_CASES = [
    # Environment doesn't provide enough arguments
    (None, {}, pytest.raises(TypeError)),
]


@pytest.mark.parametrize("cli_args, env_mapping, context", EMPTY_TEST_CASES)
def test_threshold_empty(
    monkeypatch,
    cli_args,
    env_mapping,
    context,
):
    @threshold
    def show_hierarchy_empty(first: int, second: str, third: float = 0.0):
        return first, second, third

    if isinstance(cli_args, list) and all(isinstance(x, str) for x in cli_args):
        monkeypatch.setattr("sys.argv", cli_args)

    else:
        monkeypatch.setattr("sys.argv", ["None"])

    with monkeypatch.context() as m, context:
        for env_var_name, env_var_value in env_mapping.items():
            m.setenv(env_var_name, env_var_value)

        show_hierarchy_empty()


PARAMETRIZED_TEST_CASES = [
    # Default empty, recreated for parity
    ("", {}, "", "", ENV_PREFIX, CLI_PATTERN, "", False, pytest.raises(TypeError)),
    # Add tracing with correct log level - WARNING
    (
        "",
        {},
        "",
        "",
        ENV_PREFIX,
        CLI_PATTERN,
        "WARNING",
        False,
        pytest.raises(TypeError),
    ),
    # Add tracing with incorrect log level - TURING
    (
        "",
        {},
        "",
        "",
        ENV_PREFIX,
        CLI_PATTERN,
        "TURING",
        False,
        pytest.raises(ValueError),
    ),
    # Add fpath_json only
    (
        "",
        {},
        "tests/json/good_config.json",  # tests are run from base directory
        "",
        ENV_PREFIX,
        CLI_PATTERN,
        "",
        False,
        no_error,
    ),
    # Add fpath_yaml to previous - also debug = True
    (
        "",
        {},
        "tests/json/good_config.json",  # tests are run from base directory
        "tests/yaml/good_config.yaml",
        ENV_PREFIX,
        CLI_PATTERN,
        "SUCCESS",
        True,
        no_error,
    ),
    # Add Environment Variables to previous
    (
        "",
        {"THRESH_FIRST": "99"},
        "tests/json/good_config.json",  # tests are run from base directory
        "tests/yaml/good_config.yaml",
        ENV_PREFIX,
        CLI_PATTERN,
        "SUCCESS",
        True,
        no_error,
    ),
    # Add CLI to previous
    (
        "--second=Liskov",
        {"THRESH_FIRST": "99"},
        "tests/json/good_config.json",  # tests are run from base directory
        "tests/yaml/good_config.yaml",
        ENV_PREFIX,
        CLI_PATTERN,
        "SUCCESS",
        True,
        no_error,
    ),
]


@pytest.mark.parametrize(
    "cli_args,env_mapping,fpath_json,fpath_yaml,env_prefix,"
    "cli_pattern,trace_level,debug,context",
    PARAMETRIZED_TEST_CASES,
)
def test_threshold_parametrized(
    monkeypatch,
    cli_args,
    env_mapping,
    fpath_json,
    fpath_yaml,
    env_prefix,
    cli_pattern,
    trace_level,
    debug,
    context,
):
    @threshold(
        fpath_json=fpath_json,
        fpath_yaml=fpath_yaml,
        env_prefix=env_prefix,
        cli_pattern=cli_pattern,
        trace_level=trace_level,
        debug=debug,
    )
    def show_hierarchy(first: int, second: str, third: float = 0.0):
        return first, second, third

    if isinstance(cli_args, list) and all(isinstance(x, str) for x in cli_args):
        monkeypatch.setattr("sys.argv", cli_args)

    else:
        monkeypatch.setattr("sys.argv", ["None"])

    with monkeypatch.context() as m, context:
        for env_var_name, env_var_value in env_mapping.items():
            m.setenv(env_var_name, env_var_value)

        show_hierarchy()


OVERRIDE_TEST_CASES = [
    # Normally an error, the argument is overriden at runtime by the developer
    ("", {}, "", "", ENV_PREFIX, CLI_PATTERN, "", False, no_error),
    (
        "--second=Liskov",
        {"THRESH_FIRST": "99"},
        "tests/json/good_config.json",  # tests are run from base directory
        "tests/yaml/good_config.yaml",
        ENV_PREFIX,
        CLI_PATTERN,
        "SUCCESS",
        True,
        no_error,
    ),
]


@pytest.mark.parametrize(
    "cli_args,env_mapping,fpath_json,fpath_yaml,env_prefix,"
    "cli_pattern,trace_level,debug,context",
    OVERRIDE_TEST_CASES,
)
def test_override_threshold(
    monkeypatch,
    cli_args,
    env_mapping,
    fpath_json,
    fpath_yaml,
    env_prefix,
    cli_pattern,
    trace_level,
    debug,
    context,
):
    @threshold(
        fpath_json=fpath_json,
        fpath_yaml=fpath_yaml,
        env_prefix=env_prefix,
        cli_pattern=cli_pattern,
        trace_level=trace_level,
        debug=debug,
    )
    def show_hierarchy(first: int, second: str, third: float = 0.0):
        return first, second, third

    if isinstance(cli_args, list) and all(isinstance(x, str) for x in cli_args):
        monkeypatch.setattr("sys.argv", cli_args)

    else:
        monkeypatch.setattr("sys.argv", ["None"])

    with monkeypatch.context() as m, context:
        for env_var_name, env_var_value in env_mapping.items():
            m.setenv(env_var_name, env_var_value)

        show_hierarchy(first=1, second="two", third=3.0)



"""Unit tests for the argmerge._yaml::parse_yaml() function."""

import pytest
from yaml.scanner import ScannerError

from argmerge.yaml import parse_yaml
from tests.utils import no_error

fpath_bad = "fpath_config.json"


@pytest.mark.parametrize(
    "threshold_kwargs,change_ledger,fpath_yaml,debug,expected_kwargs,expected_ledger,context",
    [
        # file does not exist
        pytest.param(
            {},
            {},
            "missing_config.yaml",
            False,
            None,
            None,
            pytest.raises(FileNotFoundError),
        ),
        # wrong filepath extension -> ValueError
        pytest.param(
            {},
            {},
            fpath_bad,
            False,
            None,
            None,
            pytest.raises(ValueError),
        ),
        # broken YAML -> error
        pytest.param(
            {},
            {},
            "tests/yaml/bad_config.yaml",
            False,
            None,
            None,
            pytest.raises(ScannerError),
        ),
        # working example
        # debug true
        pytest.param(
            {},
            {},
            "tests/yaml/good_config.yaml",
            True,
            {"first": 1, "second": {"third": 3, "d": 4}},
            {
                "first": {"label": "YAML (tests/yaml/good_config.yaml)", "rank": 20},
                "second": {"label": "YAML (tests/yaml/good_config.yaml)", "rank": 20},
            },
            no_error,
        ),
        # debug false
        pytest.param(
            {"first": 7, "third": 9},
            {
                "first": {"label": "JSON (other.json)", "rank": 10},
                "third": {"label": "JSON (other.json)", "rank": 10},
            },
            "tests/yaml/good_config.yaml",
            True,
            {"first": 1, "second": {"third": 3, "d": 4}, "third": 9},
            {
                "first": {"label": "YAML (tests/yaml/good_config.yaml)", "rank": 20},
                "second": {"label": "YAML (tests/yaml/good_config.yaml)", "rank": 20},
                "third": {"label": "JSON (other.json)", "rank": 10},
            },
            no_error,
        ),
    ],
)
def test_parse_yaml(
    threshold_kwargs,
    change_ledger,
    fpath_yaml,
    debug,
    expected_kwargs,  # output[0]
    expected_ledger,  # output[1]
    context,  # whether or not is an error
):
    with context:
        threshold_kwargs_, change_ledger_ = parse_yaml(
            threshold_kwargs, change_ledger, fpath_yaml, debug=debug
        )
        assert threshold_kwargs_ == expected_kwargs
        assert change_ledger_ == expected_ledger



import pytest

from argmerge.trace import _log_trace, _write_trace, trace_arg_lineage
from tests.utils import no_error


@pytest.mark.parametrize(
    "message,level,context",
    [
        ("Hello", "INFO", no_error),
        ("Bad level", "BAD", pytest.raises(ValueError)),
        ("WORSE LEVEL", -1, pytest.raises(TypeError)),
    ],
)
def test_write_trace(message, level, context):
    with context:
        _write_trace(message=message, level=level)


def fake_function(first: int, second: str, third: float = 0.0):
    pass


@pytest.mark.parametrize(
    "f,change_ledger,level,context",
    [
        (fake_function, {}, "INFO", no_error),
        (
            fake_function,
            {"third": {"label": "Python Function default", "rank": 0}},
            "INFO",
            no_error,
        ),
    ],
)
def test_trace_arg_lineage(f, change_ledger, level, context):
    with context:
        trace_arg_lineage(f, change_ledger, level)


@pytest.mark.parametrize("ledger,level,context", [({}, "INFO", no_error)])
def test_log_trace(ledger, level, context):
    with context:
        _log_trace(ledger, level)
